@@ -5,7 +5,7 @@ use strict;
 require Exporter;
 our ( @ISA, @EXPORT_OK, %EXPORT_TAGS );
 
-our $VERSION = 0.6;
+our $VERSION = 0.7;
 
 use IO::Select;
 use constant 1.01;
@@ -41,16 +41,7 @@ use constant XBEE_API_TYPE_TO_STRING => {
     0x95 => 'NODE_IDENTIFICATION_INDICATOR',
 };
 
-use constant XBEE_API_BAUD_RATE_TABLE => [
-    1200,
-    2400,
-    4800,
-    9600,
-    19200,
-    38400,
-    57600,
-    115200,
-];
+use constant XBEE_API_BAUD_RATE_TABLE => [1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200];
 
 use constant XBEE_API_BROADCAST_ADDR_H          => 0x00;
 use constant XBEE_API_BROADCAST_ADDR_L          => 0xFFFF;
@@ -190,6 +181,21 @@ will help work around any issues with frame ID's "leaking", but could cause odd
 behavior in cases where all outstanding frame IDs are still in use. This option
 should be used with caution.
 
+=head3 alloc_frame_id_func
+=head3 free_frame_id_func
+
+Optional code references to functions used to allocate and free frame IDs. If
+both are specified they will be called in place of the internal frame ID
+tracking functions allowing the user more control over how frame IDs are
+generated. The alloc_frame_id_func will be called when a new frame ID is needed
+and will be passed as a parameter the reference to the Device:XBee::API object
+and must return an integer between 1 and 255 inclusive. The free_frame_id_func
+will be called when the reply frame is received and the frame ID is no longer
+needed and will be passed as parameters a reference to the Device::XBee::API
+obect and the frame ID to be freed.
+
+See L<auto_reuse_frame_id> for details on how this module uses frame IDs.
+
 =head3 api_mode_escape
 
 Optional. If set to a true value, the module will automatically escape outgoing
@@ -211,6 +217,11 @@ sub new {
     $self->{auto_reuse_frame_id} = $options->{auto_reuse_frame_id} ? 1 : 0;
     $self->{api_mode_escape}     = $options->{api_mode_escape} ? 1 : 0;
 
+    if ( $options->{alloc_frame_id_func} && $options->{free_frame_id_func} ) {
+        $self->{alloc_frame_id_func} = $options->{alloc_frame_id_func};
+        $self->{free_frame_id_func}  = $options->{free_frame_id_func};
+    }
+
     $self->{in_flight_uart_frames} = {};
     $self->{known_nodes}           = {};
     $self->{rx_queue}              = [];
@@ -223,23 +234,23 @@ sub new {
     }
 
     if ( $self->{api_mode_escape} ) {
-        $self->{api_mode_escape_table} = {};
+        $self->{api_mode_escape_table}   = {};
         $self->{api_mode_unescape_table} = {};
         # Note the unescape re starts with the escape character.
-        $self->{api_mode_escape_re} = "([";
+        $self->{api_mode_escape_re}   = "([";
         $self->{api_mode_unescape_re} = "\x7D([";
         # List of characters taken from XBee datasheet.
         foreach my $e ( 0x7E, 0x7D, 0x11, 0x13 ) {
-            my $chr_e = chr( $e );
+            my $chr_e    = chr( $e );
             my $chr_e_20 = chr( $e ^ 0x20 );
-            $self->{api_mode_escape_table}->{ $chr_e } = $chr_e_20;
-            $self->{api_mode_unescape_table}->{ $chr_e_20 } = $chr_e;
-            $self->{api_mode_escape_re} .= quotemeta( $chr_e );
+            $self->{api_mode_escape_table}->{$chr_e}      = $chr_e_20;
+            $self->{api_mode_unescape_table}->{$chr_e_20} = $chr_e;
+            $self->{api_mode_escape_re}   .= quotemeta( $chr_e );
             $self->{api_mode_unescape_re} .= quotemeta( $chr_e_20 );
         }
 
         # Note the trailing "])" to terminate the character class.
-        $self->{api_mode_escape_re} = qr/$self->{api_mode_escape_re}])/;
+        $self->{api_mode_escape_re}   = qr/$self->{api_mode_escape_re}])/;
         $self->{api_mode_unescape_re} = qr/$self->{api_mode_unescape_re}])/;
     }
 
@@ -257,6 +268,9 @@ sub read_bytes {
     if ( !$self->{fh_sel} ) {
         while ( $timeout > 0 ) {
             my ( $count, $saw ) = $self->{fh}->read( $to_read );    # will read _up to_ 255 chars
+            if ( !defined $count ) {
+                die "Error reading from device: $!";
+            }
             if ( $count > 0 ) {
                 $chars += $count;
                 $buffer .= $saw;
@@ -290,30 +304,28 @@ sub read_bytes {
 sub read_packet {
     my ( $self ) = @_;
     my $d;
+    my $packet_data_length;
 
     do {
         $d = $self->read_bytes( 1 );
         return undef if !defined $d;
     } while ( $d ne "\x7E" );
 
-    $d = $self->read_bytes( 2 );
-    my ( $packet_data_length ) = unpack( 'n', $d );
-
-    $d = $self->read_bytes( $packet_data_length + 1 );
-    if ( !$d ) {
-        #warn "Read a partial packet!";
-        return undef;
-    }
-    $packet_data_length--;
-
     if ( $self->{api_mode_escape} ) {
-        # Since the unescape re starts with \x7D, it's okay to start matching
-        # at the begining of the received data. Even though the first byte is
-        # never escaped, we won't match it because we look for \x7D preceeding
-        # the escaped value, which we won't find before the first \x7E byte.
-        $d =~ s/$self->{api_mode_unescape_re}/$self->{api_mode_unescape_table}->{$1}/g;
+        ( $packet_data_length, $d ) = $self->read_escaped_packet();
+    } else {
+        $d = $self->read_bytes( 2 );
+
+        ( $packet_data_length ) = unpack( 'n', $d );
+
+        $d = $self->read_bytes( $packet_data_length + 1 );
+        if ( !$d ) {
+            return undef;
+        }
     }
 
+    $packet_data_length--;
+
     my ( $packet_api_id, $packet_data, $packet_checksum ) = unpack( "Ca[$packet_data_length]C", $d );
     my $validate_checksum = $packet_api_id + $packet_checksum;
     for ( my $i = 0; $i < $packet_data_length; $i++ ) {
@@ -328,19 +340,72 @@ sub read_packet {
     return ( $packet_api_id, $packet_data );
 }
 
+sub read_escaped_packet {
+    my ( $self ) = @_;
+
+    my $l1 = $self->read_bytes( 1 );
+    return unless defined $l1;
+
+    if ( $l1 eq "\x7D" ) {
+        $l1 = $self->read_bytes( 1 );
+        return unless defined $l1;
+        $l1 ^= "\x20";
+    }
+
+    my $l2 = $self->read_bytes( 1 );
+    return unless defined $l2;
+
+    if ( $l2 eq "\x7D" ) {
+        $l2 = $self->read_bytes( 1 );
+        return unless defined $l2;
+        $l2 ^= "\x20";
+    }
+
+    my $packet_data_length = unpack( 'n', $l1 . $l2 );
+    my $data = $self->read_bytes( $packet_data_length + 1 );
+    return unless defined $data;    # includes checksum
+
+    if ( $data =~ /\x7D$/ ) {       # trailing escape
+        my $tail = $self->read_bytes( 1 );
+        return unless defined $tail;
+        $data .= $tail;
+    }
+
+    $data =~ s/$self->{api_mode_unescape_re}/$self->{api_mode_unescape_table}->{$1}/g;
+    my $need_a_few_more = $packet_data_length - length( $data ) + 1;
+
+    while ( $need_a_few_more-- ) {
+        my $b = $self->read_bytes( 1 );
+        return unless defined $b;
+        if ( $b eq "\x7D" ) {
+            $b = $self->read_bytes( 1 );
+            return unless defined $b;
+            $b ^= "\x20";
+        }
+        $data .= $b;
+    }
+
+    return ( $packet_data_length, $data );
+}
+
 sub free_frame_id {
     my ( $self, $id ) = @_;
+    if ( $self->{free_frame_id_func} ) { return $self->{free_frame_id_func}->( $self, $id ); }
     delete $self->{in_flight_uart_frames}->{$id};
 }
 
 # id 0 is special, don't allocate it. I don't know if we should die here or
 # return 0 on failure...
 sub alloc_frame_id {
-    my ( $self )    = @_;
+    my ( $self ) = @_;
+
+    if ( $self->{alloc_frame_id_func} ) { return $self->{alloc_frame_id_func}->( $self ); }
+
     my $start_id    = int( rand( 255 ) ) + 1;
     my $id          = $start_id;
     my $oldest_time = 0xFFFFFFFF;
     my $oldest_id;
+
     while ( 1 ) {
         if ( !exists $self->{in_flight_uart_frames}->{$id} ) {
             $self->{in_flight_uart_frames}->{$id} = time();
@@ -917,16 +982,16 @@ sub __data_to_int {
 
 sub __parse_modem_status {
     my ( $api_data ) = @_;
-    my @u = unpack( 'C', $api_data );
+    my $u = unpack( 'C', $api_data );
     return {
-        status            => $u[1],
-        is_hardware_reset => $u[1] == 1,
-        is_wdt_reset      => $u[1] == 2,
-        is_associated     => $u[1] == 3,
-        is_disassociated  => $u[1] == 4,
-        is_sync_lost      => $u[1] == 5,
-        is_coord_realign  => $u[1] == 6,
-        is_coord_start    => $u[1] == 7,
+        status            => $u,
+        is_hardware_reset => $u == 1,
+        is_wdt_reset      => $u == 2,
+        is_associated     => $u == 3,
+        is_disassociated  => $u == 4,
+        is_sync_lost      => $u == 5,
+        is_coord_realign  => $u == 6,
+        is_coord_start    => $u == 7,
     };
 }
 
@@ -1144,6 +1209,12 @@ Miscellaneous code examples follow.
 
 =head1 CHANGES
 
+=head2 0.7, 20130330 - jeagle
+
+Add ability to allow users to specify their own frame allocation routines.
+
+Update API mode 2 with latest version from jdodgen
+
 =head2 0.6, 20120624 - jeagle
 
 Update documentation.
